#include "common.h"
#include "dict.h"
#include "util.h"



typedef Int16 (*DictSortFunc)(char* key1,char* val1,char* key2,char* vel2);


Dict* dictNew(){
	
	Dict* d = malloc(sizeof(Dict));
	
	if(!d) return NULL;
	
	d->numEntries = 0;
	d->keys = malloc(1);
	d->vals = malloc(1);
	
	if(d->keys && d->vals){
		
		return d;
	}
	else if(d->keys){
		
		free(d->keys);
	}
	else if(d->vals){
		
		free(d->vals);
	}
	
	free(d);
	return NULL;
}

void dictDestroy(Dict* dict){
	
	dictEmpty(dict);
	free(dict);
}

void dictEmpty(Dict* dict){
	
	UInt16 i;
	
	for(i = 0; i < dict->numEntries; i++){
		
		free(dict->keys[i]);
		free(dict->vals[i]);
	}
	
	dict->numEntries = 0;
}

void dictDel(Dict* dict,UInt16 i){
	
	if(i >= dict->numEntries) return;
	
	free(dict->keys[i]);
	free(dict->vals[i]);
	
	MemMove(dict->keys + i,dict->keys + i + 1,4 * (dict->numEntries - i - 1));
	MemMove(dict->vals + i,dict->vals + i + 1,4 * (dict->numEntries - i - 1));
	
	dict->numEntries--;
}

Boolean dictGet(Dict* dict,UInt16 i,char** key,char** val){
	
	if(i >= dict->numEntries) return false;
	
	if(key) *key = dict->keys[i];
	if(val) *val = dict->vals[i];
	
	return true;
}

void dictMod(Dict* dict,UInt16 i,char* newKey,char* newVal){
	
	if(i >= dict->numEntries) return;
	
	if(StrLen(dict->keys[i]) >= StrLen(newKey)){
		
		StrCopy(dict->keys[i],newKey);
	}
	else{
		
		free(dict->keys[i]);
		dict->keys[i] = strdup(newKey);
	}
	
	if(StrLen(dict->vals[i]) >= StrLen(newVal)){
		
		StrCopy(dict->vals[i],newVal);
	}
	else{
		
		free(dict->vals[i]);
		dict->vals[i] = strdup(newVal);
	}
}

void dictSort(Dict* dict,DictSortFunc cmp){
	
	UInt16 i,j,k;
	char* t;
	
	for(i = 0; i < dict->numEntries;i++){
		
		k = i;
		
		for(j = i + 1; j < dict->numEntries; j++){
			
			if(cmp(dict->keys[k],dict->vals[k],dict->keys[j],dict->vals[j]) < 0){
				
				k = j;
			}
		}
		
		t = dict->keys[i];
		dict->keys[i] = dict->keys[k];
		dict->keys[k] = t;
		
		t = dict->vals[i];
		dict->vals[i] = dict->vals[k];
		dict->vals[k] = t;
	}
}

Boolean dictAdd(Dict* dict,char* key,char* val,UInt16* itemIdxP){
	
	UInt16 i;
	char** na;
	UInt32 sz;
	
	
	//remove the following line if this is too slow, but then duplicate keys can exist and the OS will only see the first....
	for(i = 0; i < dict->numEntries; i++) if(StrCompare(dict->keys[i],key) == 0) return;
	
	sz = sizeof(char*) * (UInt32)(dict->numEntries);
	
	na = malloc(sizeof(char*) + sz);
	if(!na) return false;
	MemMove(na,dict->keys,sz);
	free(dict->keys);
	dict->keys = na;
	
	na = malloc(sizeof(char*) + sz);
	if(!na) return false;
	MemMove(na,dict->vals,sz);
	free(dict->vals);
	dict->vals = na;
	
	dict->keys[dict->numEntries] = strdup(key);
	dict->vals[dict->numEntries] = strdup(val);
	
	if(itemIdxP) *itemIdxP = dict->numEntries;
	
	dict->numEntries++;
	
	return true;
}

UInt16 dictCount(Dict* dict){
	
	return dict->numEntries;
}

Boolean dictLookup(Dict* dict,char* k,UInt16* idxP,char** valP){
	
	UInt16 i;
	
	for(i = 0; i < dict->numEntries; i++){
		
		if(StrCompare(k,dict->keys[i]) == 0){
			
			if(valP) *valP = dict->vals[i];
			if(idxP) *idxP = i;
			
			return true;
		}
	}
	
	return false;
}