#include "common.h"
#include "osDict.h"
#include "util.h"


typedef struct{
	
	void*	res;
	UInt16	numPairs;
	UInt8	pad2[0x0A];
	char**	srcs;
	char**	dsts;
	
}OsDict;


#define SUPER_SECRET_KEY		'HFpG'

static OsDict* getMembers();



Boolean osDictRead(Dict* intoHere){
	
	OsDict* d = getMembers();
	UInt16 i,L;
	char** ss;
	char** ds;
	
	if(!d) return false;
	
	L = swap16(d->numPairs);
	ss = swapptr(d->srcs);
	ds = swapptr(d->dsts);
	
	dictEmpty(intoHere);
	
	for(i = 0; i < L; i++){
		
		if(!dictAdd(intoHere,swapptr(ss[i]),swapptr(ds[i]),NULL)){
			
			dictEmpty(intoHere);
			return false;
		}
	}
	
	return true;
}

Boolean osDictWrite(Dict* fromHere){
	
	OsDict* d = getMembers();
	UInt16 i,L;
	char** ss;
	char** ds;
	
	if(!d) return false;
	
	if(swapptr(*(char**)swapptr(d->srcs)) >= swapptr(d->res)){	//original mode?
		
		//no cleanup to do
	}
	else{
		
		L = swap16(d->numPairs);
		ss = swapptr(d->srcs);
		ds = swapptr(d->dsts);
		
		for(i = 0; i < L; i++){
			
			free(swapptr(ss[i]));
			free(swapptr(ds[i]));
		}
		
		free(ss);
		free(ds);
		
		d->numPairs = 0;		//for safety...
	}
	
	//see how many entries there are...
	
	L = dictCount(fromHere);
	
	ss = malloc(sizeof(char*) * (UInt32)L);
	ds = malloc(sizeof(char*) * (UInt32)L);
	
	if(ss && ds){
		
		char* k;
		char* v;
		
		for(i = 0; i < L; i++){
			
			dictGet(fromHere,i,&k,&v);
			ss[i] = swapptr(strdup(k));
			ds[i] = swapptr(strdup(v));
		}
		
		d->srcs = swapptr(ss);
		d->dsts = swapptr(ds);
		d->numPairs = swap16(L);
		
		return true;
	}
	else if(ss){
		
		free(ss);
	}
	else if(ds){
		
		free(ds);
	}
	
	return false;
}

//the following functions are a mess...don't ask...

asm static void* getBase(){
	link	a6,#-4
	move.l	#0 + SUPER_SECRET_KEY,-4(a6)
	clr.l	-(a7)
	lea		-3(a6),a0
	move.l	a0,-(a7)
	dc.l	0x4E4FA45A
	unlk	a6
	rts
}

static Boolean getOsOffsets(UInt32* tableOffsetP,UInt32* itemOffsetP);

static OsDict* getMembers(){
	
	void** val = getBase();
	UInt8* ptr = NULL;
	UInt32 table,item;
	
	val = swapptr(val[0]);
	val = swapptr(val[110]);
	
	if(!getOsOffsets(&table,&item)) return NULL;
	
	ptr = swapptr(val[table / 4]);
	
	ptr += item;
	
	return (OsDict*)ptr;
}

Boolean isLdrStrConstOffset(Int32 instr,UInt32* dstRegP,UInt32* srcRegP,Int32* offsetP,Boolean* lrdP){
	
	if((instr & 0xFF600000) == 0xE5000000){
		
		Int32 offset;
		
		*dstRegP = (instr >> 12) & 0x0F;
		*srcRegP = (instr >> 16) & 0x0F;
		
		offset = instr & 0xFFF;
		if(!(instr & 0x00800000)) offset = -offset;
		
		*offsetP = offset;
		
		*lrdP = ((instr & 0x00100000) != 0);
		
		return true;
	}
	
	return false;
}

Boolean isLdrConstToReg(UInt32 instr,UInt32* regP,Int32* offsetToConstP){
	
	UInt32 srcReg;
	Boolean ldr;
	
	if(isLdrStrConstOffset(instr,regP,&srcReg,offsetToConstP,&ldr) && ldr && srcReg == 15){
			
		(*offsetToConstP) += 8;
		return true;
	}
	
	return false;
}

Boolean isAddInstr(UInt32 instr,UInt32* dstP,UInt32* src1P,UInt32* src2P){
	
	if((instr & 0xFFF00FF0) == 0xE0800000){
		
		*dstP = (instr >> 12) & 0x0F;
		*src1P = (instr >> 16) & 0x0F;
		*src2P = instr & 0x0F;
		
		return true;
	}
	
	return false;
}

Boolean isFuncCall(UInt32 instr,Int32* offsetToTargetP){
	
	if((instr & 0xFF000000) == 0xEB000000){
		
		Int32 target = (instr & 0x00FFFFFF) << 2;
		
		if(target & 0x02000000) target |= 0xFC000000;
		
		target += 8;
		
		*offsetToTargetP = target;
		
		return true;
	}
	
	return false;
}


//really, do not ask...ths is the AI to get the offsets for all treos :-)


static Boolean getOsOffsets(UInt32* tableOffsetP,UInt32* itemOffsetP){
	
	DmOpenRef db;
	Boolean ret = false;
	MemHandle mh;
	UInt32 i,j,k,sz;
	UInt32* mp;
	Boolean foundPart2 = false;
	UInt32 globPtrReg,const1,const2,rd,rs1,rs2,tmp;
	Int32 constOffset;
	
	db = DmOpenDatabaseByTypeCreator('aexo','HsEx',dmModeReadOnly);
	if(db){
		
		mh = DmGet1Resource('amdc',0);
		if(mh){
			
			sz = MemHandleSize(mh);
			mp = MemHandleLock(mh);
			
			sz -= 128;	//we do not search last 128 bytes...no point
			
			for(i = 0; i < sz >> 2; i++){
				
				
				foundPart2 = false;
				
				if(!isLdrConstToReg(swap32(mp[i + 0]),&globPtrReg,&constOffset)) continue;
				if(globPtrReg > 12 || globPtrReg == 9) continue;
				if(constOffset & 3) continue;
				constOffset = (constOffset / 4) + i;
				if(constOffset < 0 || constOffset > sz) continue;
				const1 = swap32(mp[constOffset]);
				
				if(!isAddInstr(swap32(mp[i + 2]),&rd,&rs1,&rs2)) continue;
				if(rd != globPtrReg) continue;
				if(rs1 != rd && rs2 != rd) continue;
				if(rs1 != 9 && rs2 != 9) continue;
				
				for(j = i + 2; j < i + 18; j++){
					
					Boolean foundR0 = false;
					Boolean foundR1 = false;
					Boolean load;
					
					for(k = j; k < j + 2; k++){
						
						if(!isLdrConstToReg(swap32(mp[k]),&rd,&constOffset)) break;
						if(constOffset & 3) break;
						constOffset = (constOffset / 4) + k;
						if(constOffset < 0 || constOffset > sz) break;
						tmp = swap32(mp[constOffset]);
						if(rd == 0 && tmp == 'tSTL'){
							
							foundR0 = true;
						}
						if(rd == 1 && tmp == 0x61AC){
							
							foundR1 = true;
						}
					}
					if(!foundR0 || !foundR1) continue;
					
					if(!isFuncCall(swap32(mp[j + 2]),&constOffset)) continue;
					constOffset = (constOffset / 4) + j + 2;
					if(constOffset < 0 || constOffset > sz) continue;
					if(mp[constOffset + 0] != 0x08C019E5) continue;
					if(mp[constOffset + 1] != 0xA4F19CE5) continue;
					
					if(!isFuncCall(swap32(mp[j + 5]),&constOffset)) continue;
					constOffset = (constOffset / 4) + j + 5;
					if(constOffset < 0 || constOffset > sz) continue;
					if(mp[constOffset + 0] != 0x08C019E5) continue;
					if(mp[constOffset + 1] != 0xE4F49CE5) continue;
					
					if(!isLdrStrConstOffset(swap32(mp[j + 6]),&rs2,&rs1,&constOffset,&load)) continue;
					if(!load) continue;
					if(constOffset) continue;
					if(rs1 != globPtrReg) continue;
					
					if(!isLdrStrConstOffset(swap32(mp[j + 7]),&rd,&rs1,&constOffset,&load)) continue;
					if(load) continue;
					if(constOffset < 0) continue;
					if(constOffset & 3) continue;
					if(rs1 != rs2) continue;
					if(rd != 0) continue;
					
					const2 = constOffset;
					foundPart2 = true;
					break;
				}
				if(foundPart2) break;
			}
			
			if(foundPart2){
				*tableOffsetP = const1;
				*itemOffsetP = const2;
				ret = true;
			}
			
			MemHandleUnlock(mh);
			DmReleaseResource(mh);
		}
		DmCloseDatabase(db);
	}
	
	return ret;
}




/*

//look for ARM:

LDR     Ry, =loc_44			//or another offset
...one instr
ADD     Ry, Ry, R9
... any number of instrs - we allow [0...16]
LDR     R1, =0x61AC
LDR     R0, aLtst       ; 'tSTL', (c1bc)
BL      __s_8_1a4
...2 instrs
BL      __s_8_4e4
LDR     Rx, [Ry]
STR     R0, [Rx,#0x168]			//or another offset

*/



