#include "common.h"
#include "dict.h"


void uiPrvSortDict(Dict* d);

static void setGlobDict(Dict* d){
	
	FtrSet(CRID,FTR_NUM_GLOB_DICT,(UInt32)d);
}

static void unsetGlobDict(){
	
	FtrUnregister(CRID,FTR_NUM_GLOB_DICT);
}

static Dict* getGlobDict(){
	
	Dict* d;
	
	if(errNone != FtrGet(CRID,FTR_NUM_GLOB_DICT,(UInt32*)&d)) d = NULL;
	
	return d;
}

static Boolean uiPrvValidItem(Dict* d,UInt16 i,MemHandle kH,MemHandle vH){
	
	Boolean ret = true;
	UInt16 foundIdx;
	char* k = MemHandleLock(kH);
	char* v = MemHandleLock(vH);
	
	if(dictLookup(d,k,&foundIdx,NULL) && foundIdx != i){
		
		ret = false;
	}
	
	if(!StrLen(k) || !StrLen(v)){
		
		ret = false;
	}
	
	MemHandleUnlock(kH);
	MemHandleUnlock(vH);
	
	return ret;
}

static Boolean uiModItem(Dict* d,UInt16 idx){
	
	Boolean ret = false;
	char* k;
	char* v;
	MemHandle kH,vH;
	
	if(dictGet(d,idx,&k,&v)){
		
		kH = MemHandleNew(StrLen(k) + 1);
		vH = MemHandleNew(StrLen(v) + 1);
		
		if(kH && vH){
			
			FormPtr fp;
			FieldPtr kF,vF;
			UInt16 b;
			
			StrCopy(MemHandleLock(kH),k);
			StrCopy(MemHandleLock(vH),v);
			
			MemHandleUnlock(kH);
			MemHandleUnlock(vH);
			
			fp = FrmInitForm(1001);
			
			kF = FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,1000));
			vF = FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,1001));
			
			FldSetTextHandle(kF,kH);
			FldSetTextHandle(vF,vH);
			
			do{
				
				b = FrmDoDialog(fp);
				
			}while(b == 2000 && !uiPrvValidItem(d,idx,kH,vH));
			
			FldSetTextHandle(kF,NULL);
			FldSetTextHandle(vF,NULL);
			
			dictMod(d,idx,MemHandleLock(kH),MemHandleLock(vH));
			
			MemHandleUnlock(kH);
			MemHandleUnlock(vH);
			
			MemHandleFree(kH);
			MemHandleFree(vH);
			
			FrmDeleteForm(fp);
			
			ret = true;
		}
		else if(kH){
			
			MemHandleFree(kH);
		}
		else if(vH){
			
			MemHandleFree(vH);
		}
	}
	
	if(ret) uiPrvSortDict(d);
	
	return ret;
}

static Boolean uiDelItem(Dict* d,UInt16 idx){
	
	dictDel(d,idx);
	
	uiPrvSortDict(d);
	
	return true;
}

static Boolean uiAddItem(Dict* d){
	
	UInt16 idx;
	char x = 0;
	
	if(dictAdd(d,&x,&x,&idx)){
		
		if(uiModItem(d,idx)){
			
			return true;
		}
		else{
			
			dictDel(d,idx);
		}
	}
	
	return false;
}

static Boolean uiMainFormHandler(EventPtr e){
	
	Dict* d = getGlobDict();
	FormPtr fp = FrmGetActiveForm();
	UInt16 id;
	Boolean handled = false;
	Boolean redraw = false;
	
	if(e->eType == ctlSelectEvent){
		
		Int16 sel = LstGetSelection(FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,1000)));
		id = e->data.ctlSelect.controlID;
		handled = true;
		
		if(id == 3000){			//save all
			
			writeOurDict(d);
		}
		else if(id == 3001){	//undo all
			
			readDefaultDict(d);
			redraw = true;
		}
		else if(id == 3002){	//load system defaults
			
			readSystemDict(d);
			redraw = true;
		}
		else if(id == 5000){	//add
			
			if(uiAddItem(d)){
				
				redraw = true;
			}
		}
		else if(id == 5001){	//change
			
			if(sel == noListSelection){
				
				ErrAlertCustom(0,"Select an item to perform this action",NULL,NULL);
			}
			else if(uiModItem(d,sel)){
				
				redraw = true;
			}
		}
		else if(id == 5002){	//del
			
			if(sel == noListSelection){
				
				ErrAlertCustom(0,"Select an item to perform this action",NULL,NULL);
			}
			else if(uiDelItem(d,sel)){
				
				redraw = true;
			}
		}
		else{
			
			handled = false;
		}
	}
	
	if(redraw){
		
		ListPtr L;
		L = FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,1000));
		LstSetListChoices(L,NULL,dictCount(d));
		FrmDrawForm(fp);
	}
	
	return handled;
}

static void uiMainFormListDraw(Int16 itemNum,RectangleType *bounds,char **itemsText){
	
	Dict* d = getGlobDict();
	RectangleType r;
	RectangleType b;
	char* k;
	char* v;
	char strQ[2];
	
	strQ[0] = '?';		//this form of initialization avoids making the string global and thus
	strQ[1] = 0;		// accessing it using the a5 register
	
	if(!dictGet(d,itemNum,&k,&v)){
		
		k = v = strQ;
	}
	
	WinGetClip(&r);
	
	RctCopyRectangle(bounds,&b);
	b.extent.x = 74;
	WinSetClip(bounds);
	WinDrawChars(k,StrLen(k),b.topLeft.x,b.topLeft.y);
	
	if(0){
		char x[13];
		
		StrIToA(x,itemNum);
		WinDrawChars(x,StrLen(x),b.topLeft.x + 40,b.topLeft.y);
	}
	
	
	RctCopyRectangle(bounds,&b);
	b.topLeft.x += 76;
	b.extent.x = 74;
	WinSetClip(bounds);
	WinDrawChars(v,StrLen(v),b.topLeft.x,b.topLeft.y);
	
	WinSetClip(&r);
}

static void uiMainForm(){
	
	Dict* d = getGlobDict();
	FormPtr fp;
	ListPtr L;
	ControlPtr enableAfterReset;
	
	fp = FrmInitForm(1000);
	L = FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,1000));
	
	LstSetDrawFunction(L,&uiMainFormListDraw);
	LstSetListChoices(L,NULL,dictCount(d));
	FrmSetEventHandler(fp,&uiMainFormHandler);
	FrmDoDialog(fp);
	FrmDeleteForm(fp);
}

void UI(){
	
	Dict* d = dictNew();
	
	if(d){
		
		readDefaultDict(d);	
		setGlobDict(d);
		uiMainForm();
		unsetGlobDict();
		dictDestroy(d);
	}
}

static Int16 prvDictSortFunc(char* key1,char* val1,char* key2,char* val2){
	
	Int16 ret;
	
	ret = StrCompare(key2,key1);
	if(ret == 0) ret = StrCompare(val2,val1);
	
	return ret;
}

static void uiPrvSortDict(Dict* d){
	
	dictSort(d,&prvDictSortFunc);
}

void readSystemDict(Dict* d){
	
	if(osDictRead(d)){
		
		uiPrvSortDict(d);
	}
	else{
		
		ErrAlertCustom(0,"Failed to read the system dictionary.",NULL,NULL);
	}
}

void readDefaultDict(Dict* d){
	
	if(dbDictRead(d) || osDictRead(d)){
		
		uiPrvSortDict(d);
	}
	else{
		
		ErrAlertCustom(0,"Failed to read any dictionary.",NULL,NULL);
	}
}

void writeOurDict(Dict* d){
	
	if(!dbDictWrite(d)){
		
		ErrAlertCustom(0,"Failed to save the dictionary to the database.",NULL,NULL);
	}
}